%VIDSYNC will syncronize some formats of video data (currently mostly only 
% AVI) with some other time stamped set of data (likely force plate data).  
% The user can select the 2 video and data files and also which frames to
% syncronize from using the VIDSYNC_VIDGUI.M gui script .
% This program will create a MATLAB movie file (exported AVI) that that 
% displays the video and other data set together.  This file has a standard
% name and is time dated.
% 04/29/05 - A. McClung - Stanford University (1st stable version)
% 04/28/06 - V41 revised specifically to use SWRI forces for plots 
%           (traction, normal, lateral)

%Clean up the workspace - either do this at start or cleanup at end
clear;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Variables that may need to be changed with video & data properties
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%NOTE: Some of these initial variables are set in both VIDSYNC_Rxx.M and VIDSYNC_VIDGUI.M, so be sure to change both.
datasamp=1000/3;                % The default SwRI sampling rate is 3ms (1000/3Hz)
framerate=30;                   % The frame rate for the exported video (Matlab default is 15fps)
quality=75;                     % The quality of the exported video (Matlab dfault is 75)
codec='Indeo5';                 % The compression method for the exported video (Matlab default is 'Indeo5')
datacols=[1 2 3 4 5 6];         % Columns containing data for plotting (All 6 legs)
syncplotzero=0;                 % Start the syncronized result at zero or actual time from input files
                                % Set to '1', to plot relatively from t=0
                                % Clear to '0', to plot in time relative to input video/data timestamps
axType='Force';                 % Type of data
axYUnits='[N]';                 % Units for Y axis
axcolor='bgrcmk';               % Colors for data plots
maxvideoframes=100;             % "Chunk size" - Maximum number of video frames loaded into memory at any time (600+ often causes virtual memory errors)
legord='031425';                % Sequential ordered leg numbers for leg plots

%Get the video data file location
[vfile,vpath]=uigetfile('*.avi','Select the video file');
%Try to automatically find the matching data file
s=strfind(vfile,'.mat');
datafound=0;
if isempty(s)==0
    %Check to see if matching file is available
    if exist([vpath vfile(1:s+3)])==2
        if menu('Load data file of same name?','Yes','No')==1
            apath=vpath;
            afile=vfile(1:s+3);
            datafound=1;
        end
    end
end    
if datafound==0
    %Get the additional (ie. force plate) data file location
    [afile,apath]=uigetfile('*.dat;*.mat','Select the data file');
end

%Determine the frame rate and length
movinfo=aviinfo([vpath vfile]);
movlength=movinfo.NumFrames/movinfo.FramesPerSecond;
fprintf('Video file length: %1.2f secs\n', movlength);
fprintf('Total frames: %d\n', movinfo.NumFrames);
%Load the video file (OR Just load 1st chunk if video is too long)
if (movinfo.NumFrames>maxvideoframes)
    frames2load=[1:maxvideoframes];
else
    frames2load=[1:movinfo.NumFrames];
end
fprintf('Loading frames [1:%d] to sync ....\n',frames2load(size(frames2load,2)));
mov2sync=MYaviread([vpath vfile],frames2load);
fprintf('DONE!\n');

%Load the data file
fprintf('Reading the data file...');
data2sync=load([apath afile]);
fprintf('\nDONE!\n');
%Determine if the data is in a struct, if so extract data, time and sampling rate
if isstruct(data2sync)==1
    structdata=1;                   % Marker for data in struct form
    fprintf('NOTE: Struct data detected!\n')
    getstructdata_SWRI;             % Script to extract the (SWRI force) data from struct 
end
% Determine the length of the data (using the traction data)
datanum=size(data2sync.trac,1);     % Look at the TRACTION data to determine length
datalength=datanum/datasamp;
fprintf('Data file length: %1.2f secs\n', datalength);

%Syncronize the video with the additional data
%  NOTE: Usually the data sampling rate will be higher than the video frame
%   rate, so the data will be synced to the video by default
%  NOTE: If the video is sampled faster than data, an ELSE statement can be
%   added to the followinig statement to handle that case.
if datasamp>=movinfo.FramesPerSecond
    %Select the first frame from video & data to start syncronizing.  The 
    % two frames are returned from the VIDSYNC_VIDGUI script.
    fprintf('Starting syncronizer GUI...\n')
    %Get the 1st (matching) VIDEO and DATA frames to sync from gui scrpt
    [movframe1 dataframe1 tsyncinfo]=vidsync_vidgui(mov2sync,data2sync,datasamp,[vpath vfile]);
    %Determine how much time from each (video & data) can be syncronized
    timemarks(1,:)=[movframe1 movinfo.NumFrames]/movinfo.FramesPerSecond;
    timemarks(2,:)=[dataframe1 datanum]/datasamp;
    synctime=min(timemarks(:,2)-timemarks(:,1));    % Time length of sync
    if (timemarks(1,2)-timemarks(1,1))>(timemarks(2,2)-timemarks(2,1))
        %Video is longer than data
        fprintf('The VIDEO footage is LONGER than the acquired DATA!\n')
        timemarks(1,2)=timemarks(1,1)+synctime;
    else
        %Data is longer than video
        fprintf('The acquired DATA is LONGER than the VIDEO footage!\n')
        timemarks(2,2)=timemarks(2,1)+synctime;
    end
    %Check for special timed sync specific variable (This allows the user
    % to only sync for a specified time period)
    if (tsyncinfo(1)==1)
        %Find the time difference between video and data (relative to video)
        toffset=(dataframe1/datasamp)-(movframe1/movinfo.FramesPerSecond);
        synctime=abs(diff(tsyncinfo(2:3)));     %Time length of sync output
        %Use the timing of the video for sync window
        if tsyncinfo(4)==1
            %Get the 1st video frame to sync
            tmovframe1=round(tsyncinfo(2)*movinfo.FramesPerSecond);
            %Use the input time and time offset to figure out the 1st data frame
            tdataframe1=round((tsyncinfo(2)+toffset)*datasamp);
            timemarks(1,:)=[tsyncinfo(2:3)];
            timemarks(2,:)=timemarks(1,:)+toffset;
            tb='Video';
        %Use the timing of the data for sync window
        else
            %Get the 1st data frame to sync
            tdataframe1=round(tsyncinfo(2)*datasamp);
            %Use the input time and time offset to figure out the 1st video frame
            tmovframe1=round((tsyncinfo(2)-toffset)*movinfo.FramesPerSecond);
            timemarks(2,:)=[tsyncinfo(2:3)];
            timemarks(1,:)=timemarks(2,:)-toffset;
            tb='Data';
        end
        %Give some feedback
        fprintf('Timed sync has been requested for %4.2f to %4.2fsec based on %s timing!\n',...
            tsyncinfo(2), tsyncinfo(3), tb)
        %Copy the new starting frames (for timed sync)
        movframe1=tmovframe1;
        dataframe1=tdataframe1;
    end
    
    %Display the video and data that will be used
    fprintf(['The video and data can be syncronized for ' num2str(synctime) ' secs!\n']);
    fprintf(['Video [Start : Stop] - TIME(s)->[ %1.4f : %1.4f ] - FRAMES->[%6d : %6d ]\n'],...
        timemarks(1,1),timemarks(1,2),movframe1,round(timemarks(1,2)*movinfo.FramesPerSecond))
    fprintf(['Data  [Start : Stop] - TIME(s)->[ %1.4f : %1.4f ] - FRAMES->[%6d : %6d ]\n'],...
        timemarks(2,1),timemarks(2,2),dataframe1,round(timemarks(2,2)*datasamp))
    fprintf('Note: Times and frames are based on original input files.\n');
    %Build synced data and remove excess (RELATIVE TO VIDEO TIMING)
    syncnum=round(timemarks(1,2)*movinfo.FramesPerSecond)-movframe1;    %Number of video frames to syncronize
    for i=1:syncnum
        %Get the video time relative to specified 1st frame
        %NOTE: EITHER '-1' or '-0' depending on where you want video to start.
        %  '-1' starts at zero, '-0' starts at 1st frame time
        synctime(i)=(i-1)/movinfo.FramesPerSecond;
        %Use the video timing to get the related data frame
        % -This the closest matching data sample for video, shifted by the specified 1st data frame
        syncdataint(i)=dataframe1+round(synctime(i)*datasamp);
        %Store the data value that is now synced to the video
    end
    if syncplotzero==1
        % Plot relatively from t=0
        syncoffset=0;
        smoothtime=([syncdataint(1):syncdataint(syncnum)]-dataframe1)/datasamp;
    else
        % Plot in time relative to input video/data
        syncoffset=dataframe1/datasamp;
        smoothtime=[dataframe1:syncdataint(syncnum)]/datasamp;
    end
    synctime=synctime + syncoffset;
end

%%% - LETS START MAKING A MATLAB MOVIE - %%%
%NOTE: The movie is generated frame by frame on the screen, and the present 
% figure is captured at each stage and added to the movie

%Get the current time, to append to filename
curtime=clock;
%Add the date...
suffix=['_' num2str(curtime(1)-2000,'%02d') num2str(curtime(2),'%02d') num2str(curtime(3),'%02d')];
%Now, add the time
suffix=[suffix '-' num2str(curtime(4),'%02d') num2str(curtime(5),'%02d') num2str(floor(curtime(6)),'%02d')];

try
    %%Create the movie framing
    h=figure;
    set(h,'Position',[10 50 500 800]);          % Screen size and placement while the movie is being generated
    M=avifile(['VidDataSync' suffix '.avi']);   % Open a movie file for writing
    M.Fps=framerate;                            % Set the frame rate
    M.Quality=quality;                          % Set the movie quality
    M.Compression=codec;                        % Set the compression method

    %Draw the initial figure
    %NOTE: May need to adjust array size for different size video frames/data plots
    arraysize=size(datacols,2)+6;               % Let video take up 3 rows of subplots (ie. 6 for 6x2 array)
    h1a=subplot(arraysize/2,2,1);
    h1=image(mov2sync(1).cdata);
    %Add a properly formatted title
    if isempty(s)==0
        %Take care of tex formatting for '_' character
        s=strfind(vfile,'_');
        vfiletex=vfile;
        %Add a '\' before each '_'
        for i=1:length(s)
            vfiletex=strcat(vfiletex(1:(s(i)-1+(i-1))),'\',vfile(s(i):length(vfile)));
        end
        %Get the location for the end of the file title
        s=strfind(vfiletex,'.mat');
        title({'Syncronized Video & Data',vfiletex(1:s-1)},'FontSize',12)
    else
        title('Syncronized Video & Data','FontSize',14)
    end
    axis image off;
    %Try to resize the image to appropriate portions
    %SEE BELOW
    for i=7:arraysize       %First 6 plots are covered by image
        eval(['h' num2str(i) '=subplot(' num2str(arraysize/2) ',2,' num2str(i) ');']);
        %Plot the real data in subplot
        eval(['plot(smoothtime,data2sync.trac(syncdataint(1):syncdataint(syncnum),' num2str(datacols(i-6)) '),''' axcolor(1) ''');'])
        hold on
        eval(['plot(smoothtime,data2sync.norm(syncdataint(1):syncdataint(syncnum),' num2str(datacols(i-6)) '),''' axcolor(2) ''');'])
        eval(['plot(smoothtime,data2sync.late(syncdataint(1):syncdataint(syncnum),' num2str(datacols(i-6)) '),''' axcolor(3) ''');'])
        axis tight; hold on;
        eval(['title(''Leg #' legord(i-6) ''')'])
        grid on;
        %Add the X-axis label for the last plot
        if (i==arraysize||i==(arraysize-1))
            xlabel('Time [sec]');
        end
        ylabel([axType ' ' axYUnits]);
        %Determine axes for data framing
        dataaxes(i,:)=axis;
        eval(['h' num2str(i) 'b=plot([0 0]+syncoffset,[-3e3 3e3],''k'',''LineWidth'',2);'])
        axis(dataaxes(i,:));      %Reset the axis after time bar plot
        %Add a (too big) legend to one data plot
        if i==7
            hL=legend('Traction','Normal','Lateral',0);
            set(hL,'FontSize',5)
        end
    end
    %Normalize all plots
    normaxis=[dataaxes(7,1:2) min(dataaxes(:,3)) max(dataaxes(:,4))];
    for i=7:arraysize
        eval(['subplot(' num2str(arraysize/2) ',2,' num2str(i) ');']);
        axis(normaxis)
    end

    %Resize and reposition the video plot
    %NOTE: This may need to occur after the other subplots are created - it was not happy above
    %NOTE: May need to adjust these values for different size video frames and number of data columns
    % Format - [lower left corner X, lower left corner Y, width, height]
    %See 'help subplot' for info on 'Position' variable
    set(h1a,'Position',[.05 .3 .9 .9])


    % Loop thru and add a chunk of video (frames) to te movie each time
    %for i=1:((movframe1+syncnum-1)/maxvideoframes)+1
    for i=1:(syncnum/maxvideoframes)+1
        %Set the frames that will be loaded for addition to the video
        frames2load=[((i-1)*maxvideoframes+movframe1):(i*maxvideoframes+movframe1)];
        %Get the correct number of frames for the first and last chunks
        if i==1
            frames2load=[movframe1:maxvideoframes];
            %Make sure frames2load is not empty for time sync beyond preview
            if (movframe1>maxvideoframes)
                frames2load=[movframe1:min(round(timemarks(1,2)*movinfo.FramesPerSecond)-1,movframe1+maxvideoframes-1)];
            end
        end
        %Get the last block right
        if (i==(syncnum/maxvideoframes)+1)
            frames2load=[frames2load(1):movframe1+syncnum-1];
        end
        %Load the necessary video chunk (except in the 1st pass)
        if (frames2load(1)~=movframe1) || (i==1&tsyncinfo(1)==1)
            clear mov2sync;
            fprintf('\nLoading frames [%d:%d] ....\n',frames2load(1),frames2load(size(frames2load,2)));
            mov2sync=MYaviread([vpath vfile],frames2load);
            fprintf('DONE!\n');
        end
        %LOOP the following to add each sync'd video & data chunk to movie
        fprintf('Syncronizing frames [%d:%d]\n',frames2load(1),frames2load(size(frames2load,2)));
        for j=frames2load
            %%Display the image
            set(h1,'CData',mov2sync((j-movframe1+1)-(i-1)*maxvideoframes).cdata)
            %%Display the additional data (Move the time bar)
            for k=7:arraysize
                eval(['set(h' num2str(k) 'b,''XData'',[1 1]*(synctime(j-movframe1+1)))'])
            end
            %%Add the frame to the movie - Based on 'help avifile'
            M=addframe(M,getframe(h));
            %Give the user some feedback
            if (mod(j,10)==0)
                fprintf([' ' num2str(j)]);
            end
            if (mod(j,100)==0)
                fprintf('\n');
            end
        end
    end

    fprintf(' ALL DONE!!\n');

    %Finalize the movie file
    M=close(M);

    %Close the figure window
    close(h);
catch
    fprintf(['\nError at frame #' num2str(j) '!!\nClosing the video file...'])
    %Finalize the movie file
    M=close(M);
    fprintf('DONE!!\n');

    %Close the figure window
    close(h);
end